\
\                               M I N M O S
\                  Minimal microcontroller operating system
\
\ MINMOS is a small system for the MTV virtual machine on AVR microcontrollers
\ It's written in MTVFC, the second Forth compiler for MTV - which itself is
\ an extention to IsForth, http://isforth.clss.net
\
\ Written by Robert stling
\ robost@telia.com (robost86@hotmail.com)
\ http://robert.zizi.org/
\

\ Word header.
\
\ 00  dword  Pointer to last header.
\ 04  dword  Pointer to code.
\ 08  byte   Flags (see below).
\ 09  byte   Name length.
\ 0a         Name.
\
\ Flags.
\
\ Bit 1      1 = immediate, 0 = normal word.

fload fc.f


0 var kernel-last

1 const flag-immediate
0 const flag-none

: kernel" ( immediate pointer -- )
  offset                               \ Save current offset.
  kernel-last compile-4                \ Compile pointer to last header.
  !> kernel-last                       \ This word's offset -> kernel-last.
  compile-4                            \ Compile pointer to word.
  compile-1                            \ Compile immediate flag.
  " ;                                  \ Compile name.

0 syscall init
1 syscall poll
2 syscall receive
3 syscall send
4 syscall freeram
5 syscall sync-flash
6 syscall exit-vm

start ]]

^$00 constant false
^$ff constant true

^$00 constant mode-interpret
^$01 constant mode-compile

var4 ram
var4 last
var4 dp
var4 input-sp

var mode
var tmp1
var tmp2
var tmp3

ram-create word-buffer ^$20 ram-allot
ram-create line-buffer ^$50 ram-allot
ram-create input-stack ^$10 ram-allot
ram-create input-stackend

create ansi-clear " 2J"
create ansi-home  " 1;1H"
create ansi-back  " 1D"

: emit   ( b1 -- ) send ;
: space  ( -- )  $20 emit ;
: key    ( -- b1 ) receive ;
: print  ( a1 -- )  c@+ { is0 adrop drop ; then 1 -  c@+ emit }
: cr     ( -- )  $0a emit ;
: escape ( -- )  $1b emit $5b emit ;
: clear  ( -- )  escape ansi-clear print ;
: home   ( -- )  escape ansi-home print ;
: back   ( -- )  escape ansi-back print ;

flag-none ' emit kernel" emit"
flag-none ' key kernel" key"
flag-none ' cr kernel" cr"

: input-push  input-sp 4 4-1 4dup -> input-sp  >a ! adrop ;
: input-drop  input-sp 4 4+1 -> input-sp ;
: input-ptr  input-sp >a @ adrop ;
: input-setptr  input-sp >a ! adrop ;
: input-empty?  input-sp input-stackend a> 4=4 ;

create str-hex " 0123456789abcdef"

: .digit ( b1 -- )  0 swap 0 swap 0 swap  str-hex a> 4+4 1 4+1 >a  c@ adrop
  emit ;
: .1     ( b1 -- )  dup 2/ 2/ 2/ 2/ .digit  $0f and goto .digit
: .4     ( w1 -- )  -> tmp1  -> tmp2  -> tmp3
                    .1 tmp3 .1 tmp2 .1 tmp1 goto .1

: ipbase+ ( a1 -- a1+ipbase ) ia# 0  a> a> 4+4 >a ;
: ipbase ( -- w1 ) ia# 0 a> ;
: a- ( a1 -- a1-1 )  a> 1 4-1 >a ;
: is-space? ( b1 -- b1 )  dup $20 =  swap $0a =  or ;
: 4abs ( w1 -- w2 )  4dup 4<0 if 4neg then ;

: wait-char ( a1 -- a2 b1 )
  { c@+ dup is-space?  if0 ; then  drop }

: c, ( b1 -- )
  dp >a c!+ a> -> dp ;
: , ( w1 -- )
  dp >a ! a> 4 4+1 -> dp ;

: handle-char ( a1 a2 b1 -- a1 a2 )
  dup $0a = if
    drop 0 c! adrop c! adrop           \ Store count, zero-terminate.
    r> 4drop space ;                   \ Return from caller.
  then
  dup $08 = if
    aover a> 1 4+1 adup a> 4=4 if0     \ Are we not on the first char?
      a- back                          \ Back one step
    then ;
  then  dup emit c!+ ;

: read-line ( a1 -- )
  cr adup 0  dup c!+                   \ Save ptr to count byte, 1 = count.
  {
    key handle-char 1 +                \ Read and process character.
  }

: found-zero
  drop adrop  aswap c! adrop           \ 0? Store count byte.
  a> 1 4-1 ;                           \ Make the buffer point at the 0.
        
: read-word ( a1 a2 -- w1 )            \ Read a2 from buffer a1, w3 = new buff.
  aswap aover 0  dup c!+               \ Save pointer to count byte, count = 0
  aswap wait-char aswap                \ Read first byte from buffer.
  is0  goto found-zero  then           \ 0? Do stuff.
  c!+ {                                \ Store first byte.
    1 +                                \ Increase count.
    aswap c@+ aswap                    \ Get next byte.
    dup is-space? if
      drop adrop aswap c! adrop a> ;   \ If space, end string.
    then
    is0  goto found-zero  then         \ If 0, end string and stuff..
    c!+                                \ Store current character.
  }

: check-input ( -- )
  input-ptr >a c@ adrop if0            \ Have we reached end of input buffer?
    input-drop                         \ Drop this input source.
    input-empty? if                    \ No more input sources?
      line-buffer adup read-line       \ Read in a line.
      a> 1 4+1 input-push              \ And let it be our new input source!
    then
  then ;

: get-word ( -- )
  check-input                          \ Check if there is any more data.
  input-ptr >a word-buffer read-word   \ Read a word.
  input-setptr ;                       \ Update pointer.

: kernel-readbyte ( -- b1 )            \ Read byte from input source.
  check-input input-ptr >a             \ Get input pointer.
  c@+ a> input-setptr ;                \ Get byte, return

: kernel-" ( -- )
  dp >a adup                           \ Two copies of adup on the addr stack.
  c@+ drop 0                           \ Add 1 to top of addr stack, count=0.
  {
    kernel-readbyte dup $27 = if       \ Single quote?
      a> -> dp                         \ Set dp to end of string.
      drop c! adrop ;                  \ Store count.
    then c!+                           \ Store character.
  1 + } ;

: save-all ( -- )
  sync-flash                           \ Write Flash buffer.
  dp a# $0100000a ! adrop ;            \ Write pointer to free heap space.


create str-runtime " Error: Trying to interpret a compile-time word - "
: error-runtime  str-runtime print  word-buffer print cr ;

create str-jump " Error: Jump too long!"
: error-jump  str-jump print cr ;

create str-unknown " Error: Uknown word - "
: error-unknown  str-unknown print  word-buffer print cr ;

create str-number " Error: Invalid number - "
: error-number  str-number print  word-buffer print cr ;

: kernel-bye  save-all exit-vm ;
\ : kernel-bye  save-all  4# $01000000 ^0 jumpf
flag-none ' kernel-bye kernel" bye"

: kernel-drop  mode mode-compile = if  $00 c, ;  then drop ;
flag-immediate ' kernel-drop kernel" drop"
: kernel-dup  mode mode-compile = if  $01 c, ;  then dup ;
flag-immediate ' kernel-dup kernel" dup"
: kernel-over  mode mode-compile = if  $02 c, ;  then over ;
flag-immediate ' kernel-over kernel" over"
: kernel-swap  mode mode-compile = if  $03 c, ;  then swap ;
flag-immediate ' kernel-swap kernel" swap"

: kernel-4drop  mode mode-compile = if  $04 c, ;  then 4drop ;
flag-immediate ' kernel-4drop kernel" 4drop"
: kernel-4dup  mode mode-compile = if  $05 c, ;  then 4dup ;
flag-immediate ' kernel-4dup kernel" 4dup"
: kernel-4over  mode mode-compile = if  $06 c, ;  then 4over ;
flag-immediate ' kernel-4over kernel" 4over"
: kernel-4swap  mode mode-compile = if  $07 c, ;  then 4swap ;
flag-immediate ' kernel-4swap kernel" 4swap"

: kernel-adrop  mode mode-compile = if  $08 c, ;  then adrop ;
flag-immediate ' kernel-adrop kernel" adrop"
: kernel-adup  mode mode-compile = if  $09 c, ;  then adup ;
flag-immediate ' kernel-adup kernel" adup"
: kernel-aover  mode mode-compile = if  $0a c, ;  then aover ;
flag-immediate ' kernel-aover kernel" aover"
: kernel-aswap  mode mode-compile = if  $0b c, ;  then aswap ;
flag-immediate ' kernel-aswap kernel" aswap"

: kernel-+  mode mode-compile = if  $0c c, ;  then + ;
flag-immediate ' kernel-+ kernel" +"
: kernel--  mode mode-compile = if  $0d c, ;  then - ;
flag-immediate ' kernel-- kernel" -"
: kernel-and  mode mode-compile = if  $0e c, ;  then and ;
flag-immediate ' kernel-and kernel" and"
: kernel-or  mode mode-compile = if  $0f c, ;  then or ;
flag-immediate ' kernel-or kernel" or"
: kernel-xor  mode mode-compile = if  $10 c, ;  then xor ;
flag-immediate ' kernel-xor kernel" xor"
: kernel-not  mode mode-compile = if  $11 c, ;  then not ;
flag-immediate ' kernel-not kernel" not"
: kernel-neg  mode mode-compile = if  $12 c, ;  then neg ;
flag-immediate ' kernel-neg kernel" neg"
: kernel-2*  mode mode-compile = if  $13 c, ;  then 2* ;
flag-immediate ' kernel-2* kernel" 2*"
: kernel-2/  mode mode-compile = if  $14 c, ;  then 2/ ;
flag-immediate ' kernel-2/ kernel" 2/"

: kernel-4+1  mode mode-compile = if  $15 c, ;  then 4+1 ;
flag-immediate ' kernel-4+1 kernel" 4+1"
: kernel-4-1  mode mode-compile = if  $16 c, ;  then 4-1 ;
flag-immediate ' kernel-4-1 kernel" 4-1"
: kernel-4neg  mode mode-compile = if  $17 c, ;  then 4neg ;
flag-immediate ' kernel-4neg kernel" 4neg"
: kernel-<<  mode mode-compile = if  $18 c, ;  then << ;
flag-immediate ' kernel-<< kernel" <<"
: kernel->>  mode mode-compile = if  $19 c, ;  then >> ;
flag-immediate ' kernel->> kernel" >>"
: kernel-4+4  mode mode-compile = if  $1a c, ;  then 4+4 ;
flag-immediate ' kernel-4+4 kernel" 4+4"
: kernel-4-4  mode mode-compile = if  $1b c, ;  then 4-4 ;
flag-immediate ' kernel-4-4 kernel" 4-4"
: kernel-4and  mode mode-compile = if  $1c c, ;  then 4and ;
flag-immediate ' kernel-4and kernel" 4and"
: kernel-4or  mode mode-compile = if  $1d c, ;  then 4or ;
flag-immediate ' kernel-4or kernel" 4or"
: kernel-4xor  mode mode-compile = if  $1e c, ;  then 4xor ;
flag-immediate ' kernel-4xor kernel" 4xor"

: kernel-=  mode mode-compile = if  $1f c, ;  then = ;
flag-immediate ' kernel-= kernel" ="
: kernel-<>  mode mode-compile = if  $20 c, ;  then <> ;
flag-immediate ' kernel-<> kernel" <>"
: kernel-<  mode mode-compile = if  $21 c, ;  then < ;
flag-immediate ' kernel-< kernel" <"
: kernel->  mode mode-compile = if  $22 c, ;  then > ;
flag-immediate ' kernel-> kernel" >"

: kernel-4=1  mode mode-compile = if  $23 c, ;  then 4=1 ;
flag-immediate ' kernel-4=1 kernel" 4=1"
: kernel-4=4  mode mode-compile = if  $24 c, ;  then 4=4 ;
flag-immediate ' kernel-4=4 kernel" 4=4"
: kernel-4=0  mode mode-compile = if  $25 c, ;  then 4=0 ;
flag-immediate ' kernel-4=0 kernel" 4=0"
: kernel-4<0  mode mode-compile = if  $26 c, ;  then 4<0 ;
flag-immediate ' kernel-4<0 kernel" 4<0"
: kernel-4>=0  mode mode-compile = if  $27 c, ;  then 4>=0 ;
flag-immediate ' kernel-4>=0 kernel" 4>=0"

: kernel-c@  mode mode-compile = if  $28 c, ;  then c@ ;
flag-immediate ' kernel-c@ kernel" c@"
: kernel-c!  mode mode-compile = if  $29 c, ;  then c! ;
flag-immediate ' kernel-c! kernel" c!"
: kernel-c@+  mode mode-compile = if  $2a c, ;  then c@+ ;
flag-immediate ' kernel-c@+ kernel" c@+"
: kernel-c!+  mode mode-compile = if  $2b c, ;  then c!+ ;
flag-immediate ' kernel-c!+ kernel" c!+"
: kernel-@  mode mode-compile = if  $2c c, ;  then @ ;
flag-immediate ' kernel-@ kernel" @"
: kernel-!  mode mode-compile = if  $2d c, ;  then ! ;
flag-immediate ' kernel-! kernel" !"
: kernel->r  mode mode-compile = if  $2e c, ;  then goto error-runtime
flag-immediate ' kernel->r kernel" >r"
: kernel-r>  mode mode-compile = if  $2f c, ;  then goto error-runtime
flag-immediate ' kernel-r> kernel" r>"
: kernel->a  mode mode-compile = if  $30 c, ;  then >a ;
flag-immediate ' kernel->a kernel" >a"
: kernel-a>  mode mode-compile = if  $31 c, ;  then a> ;
flag-immediate ' kernel-a> kernel" a>"

: kernel-;  mode mode-compile = if  $3c c, ;  then goto error-runtime
flag-immediate ' kernel-; kernel" ;"
: kernel-sys  mode mode-compile = if  $42 c, ;  then sys ;
flag-immediate ' kernel-sys kernel" sys"

: kernel-if ( -- w1 )
  mode mode-compile = if
    $33 c, dp $00 c, ;
  then  goto error-runtime
flag-immediate ' kernel-if kernel" if"
: kernel-if0 ( -- w1 )
  mode mode-compile = if
    $32 c, dp $00 c, ;
  then  goto error-runtime
flag-immediate ' kernel-if0 kernel" if0"
: kernel-is ( -- w1 )
  mode mode-compile = if
    $35 c, dp $00 c, ;
  then  goto error-runtime
flag-immediate ' kernel-is kernel" is"
: kernel-is0 ( -- w1 )
  mode mode-compile = if
    $34 c, dp $00 c, ;
  then  goto error-runtime
flag-immediate ' kernel-is0 kernel" is0"

: kernel-{ ( -- w1 )
  mode mode-compile = if0
    goto error-runtime
  then  dp ;
flag-immediate ' kernel-{ kernel" {"

: patch-back ( w1 -- )
  4dup dp 4-4 4abs                     \ Get length of jump.
  126 4-1 4<0                          \ Jump too long?
  if0  goto error-jump  then           \ Yes, print error message.
  dp 4-4                               \ Get relative address.
  nip nip nip 2 -                      \ Convert word to byte, add 2 to it.
  $38 c,  c, ;                         \ Compile jump.

: kernel-} ( w1 -- )
  mode mode-compile = if0
    goto error-runtime
  then goto patch-back
flag-immediate ' kernel-} kernel" }"

: patch-rel ( w1 -- )
  4dup dp 4-4 4abs                     \ Get length of jump.
  126 4-1 4<0                          \ Is it above 126?
  if0  goto error-jump  then           \ Yes, print error message.
  >a  dp 4over 4-4 5 4-1  nip nip nip  \ Get the 8-bit relative address.
  c! adrop ;                           \ Write it to old location.
: kernel-then ( w1 -- )
  mode mode-compile = if
    patch-rel ;
  then  goto error-runtime
flag-immediate ' kernel-then kernel" then"

: digit? ( b1 -- b2 )
  dup $2f > over $3a < and             \ (x >= '0' && x <= '9')
  swap dup $60 > swap $67 < and        \ (x >= 'a' && x <= 'f')
  or ;                                 \ Either of these?
: all-digits? ( a1 -- b1 )
  c@+ c@+ drop {                       \ Get length, skip $
    1 -                                \ Remember we skipped the '$' ? :D
    is0  drop true adrop ;  then       \ End of string? Yes, return true.
    c@+ digit?                         \ Is this a hex digit?
    if0  drop false adrop ;  then      \ No, return false.
  }
: number? ( a1 -- b1 )
  c@ 1 > if0  adrop false ;  then      \ Length at least 2 characters?
  adup c@+ drop c@ adrop $24 =
  if0  adrop false ;  then             \ And first char = '$'?
  all-digits? ;                        \ And all characters are hex digits?
: char>n ( b1 -- b2 )
  dup $61 < if                         \ Below 'a'? (i.e. 0-9)
    $30 - ;                            \ Subtract '0' and return.
  then $57 - ;                         \ Otherwise subtract 'a' and return.

: (s>n) ( a1 -- w1 )                   \ Like s>n, but no error checking.
  c@+ 4# 0                             \ Get length of string and push nr.
  >r                                   \ Number on register stack.
  c@+ drop                             \ Skip $
  { 1 -                                \ Grrr! Remeber we skipped '$'? :D
    is0  drop r> adrop true ;  then    \ End of string? Yes, return number.
    r> 4 <<                            \ Shift one digit up.
    c@+ char>n 4+1 >r                  \ Convert character to number, add to n
  }    

: s>n ( a1 -- [w1] b1 )                \ b1 false means not a number, so no w1
  adup number? if0
    adrop false ;                      \ No number, return false.
  then goto (s>n)

: number ( -- [w1] b1 )
  word-buffer adup read-word s>n ;     \ Read word and convert it.
flag-none ' number kernel" number"
: kernel-4>1 ( w1 -- b1 )
  nip nip nip ;
flag-none ' kernel-4>1 kernel" 4>1"

: kernel-4# ( -- [w1] )
  number if0                           \ Read in a number.
    goto error-number                  \ On error, print message and return.
  then
  mode mode-compile = if               \ In compilation mode?
    $3f c, , ;                         \ Compile 4lit.
  then ;                               \ Otherwise, leave word on stack.
flag-immediate ' kernel-4# kernel" 4#"

: c= ( a1 a2 -- a1+1 a2+1 b1 )  c@+ aswap c@+ aswap = ;
: s=true  drop adrop adrop true ;
: s=false  drop adrop adrop false ;
: s= ( a1 a2 -- b1 )
  c@                                   \ Length of a1.
  c= if0  goto s=false  then           \ Are the strings equally long?
  {                                    \ Yes, go through every char of them.
    is0  goto s=true  then             \ End of string? Then strings are equal.
    c= if0  goto s=false  then         \ Differ? Strings not equal.
  1 - }                                \ Compare next pair of characters.

: lookup ( a1 -- [a2] b1 )             \ a2 is only returned if b1 != false.
  last >a                              \ Push address of last word.
  {
    adup a> 4=0                        \ Have we reached the end of the list?
    if  adrop adrop false ;  then      \ Yes. 2 * adrop, return false.
    ipbase+                            \ Add ipbase to address.
    aover aover a> 9 4+1 >a  s=        \ Get pointer to name string, compare.
    if  aswap adrop true ;  then       \ Equal? Then pop name and return true.
    @ adrop >a                         \ Get pointer to next word header.
  }

\ a2 = source, a1 = destination.
: scopy ( a1 a2 -- )
  c@+ aswap dup c!+ aswap              \ Get length and copy length byte.
  {
    is0  drop adrop adrop ;  then      \ End of string? Then return.
    c@+ aswap c!+ aswap                \ Copy byte.
  1 - }

: (create) ( -- )
  get-word                             \ Read word to buffer.
  dp ipbase 4-4 4dup last ,            \ Write pointer to last header.
  -> last                              \ Last points to this word.
  word-buffer c@ 4+1 $0a 4+1 ,         \ Write pointer to code.
  0 c,                                 \ No flags.
  dp 4dup  >a aswap scopy              \ Copy name string.
  word-buffer c@ adrop 1 + 4+1 -> dp ; \ Increase dp.
flag-immediate ' (create) kernel" :"
flag-none ' (create) kernel" create"

: kernel-immediate ( -- )
  last ipbase 4+4 $08 4+1 >a           \ Last word's flags byte to addr stack
  flag-immediate lit c! adrop ;        \ Set it to immediate.
flag-immediate ' kernel-immediate kernel" immediate"

label doconstant ( b1 -- [b1] )
  mode mode-compile = if               \ In compilation mode?
    $3e c, c,                          \ Compile lit.
  then ;                               \ Otherwise, just leave b1.

label doconstant4 ( w1 -- [w1] )
  mode mode-compile = if               \ In compilation mode?
    $3f c, ,                           \ Compile 4lit.
  then ;                               \ Otherwise, just leave w1.

: kernel-constant ( b1 -- )
  (create) kernel-immediate            \ Create the word.
  $3e c, c,                            \ lit b1
  $36 c, doconstant 4lit , ;           \ jump doconstant
flag-none ' kernel-constant kernel" constant"

: kernel-constant4 ( w1 -- )
  (create) kernel-immediate            \ Create the word.
  $3f c, ,                             \ lit w1
  $36 c, doconstant4 4lit , ;          \ jump doconstant4
flag-none ' kernel-constant4 kernel" constant4"

: kernel-label ( -- )
  (create) kernel-immediate            \ Create the word.
  $3f c, dp ipbase 4-4 ,               \ 4lit current offset.
  $36 c, doconstant4 4lit , ;          \ jump doconstant4
flag-none ' kernel-label kernel" label"

: kernel-[  mode-interpret -> mode ;
flag-immediate ' kernel-[ kernel" ["
: kernel-]  mode-compile -> mode ;
flag-immediate ' kernel-] kernel" ]"

: kernel-.  .1 ;
flag-none ' kernel-. kernel" ."
flag-none ' kernel-. kernel" .1"
: kernel-.4  .4 ;
flag-none ' kernel-.4 kernel" .4"

: do."
  r> ipbase 4+4 >a c@+                 \ Get length of string.
  {
    is0 drop a> ipbase 4-4 >r ; then   \ End? Return to value on addr stack.
    c@+ emit                           \ Emit one character.
  1 - }

: kernel-."
  mode mode-interpret = if
    goto error-runtime
  then
  $39 c, ' do." 4lit ,                 \ Compile call to do."
  goto kernel-"                        \ Compile the string.
flag-immediate ' kernel-." kernel" .'"

: >code ( a1 -- w1 )                   \ Return address of word code.
  a> 4 4+1 >a @ adrop ;

: call-word ( w1 -- )                  \ Call word at address w1.
  >r ;

: interpret ( a1 -- )
  mode mode-interpret = if             \ Are we in interpretation mode?
    >code call-word ;                  \ Yes, call word.
  then
  adup a> $08 4+1 >a c@ adrop          \ Address of flags.
  flag-immediate lit = if              \ Immediate?
    >code call-word ;                  \ Yes, call word.
  then
  $39 c,                               \ Compile call instruction.
  >code , ;                            \ Get address of code, compile it.

create str-logo1 "               M I N M O S"
create str-logo2 " Minimal Mincontroller Operating System"
create str-logo3 "  by Robert Oestling, robost@telia.com"

: logo ( -- )
  clear home
  str-logo1 print cr cr
  str-logo2 print cr
  str-logo3 print cr cr ;

: init-pointers ( -- )
  input-stackend a> -> input-sp        \ Pointer to input stack.
  mode-interpret -> mode               \ Not really a pointer, but...
  a# $0100000a @ adrop -> dp           \ Pointer to free heap space.
  ram-ptr 4lit -> ram                  \ Pointer to free RAM.
  kernel-last 4lit -> last ;           \ Pointer to last item in word list.

: try-word ( w1 b1 -- )
  if  interpret ;  then                \ If the word was found, interpret it.
  goto error-unknown                   \ Otherwise, show error message.

: compile-lit ( w1 -- )
  4dup $ff 4-1 4<0 if                  \ Number below 0x100? Compile lit.
    $3e c, c, drop drop drop ;         \ Compile lit.
  then
  $3f c, , ;                           \ Compile 4lit.

: try-number ( a1 -- [b1|w1] b2 )
  str-try-number print  cr
  s>n if                               \ Is this a valid number?
    mode mode-compile = if
      compile-lit true ;               \ Compile literal, return true.
    then                               \ Leave the number on stack.
    4dup $ff 4-1 4<0 if                \ Number below $100?
      nip nip nip                      \ Yes, convert to 8-bit literal.
    then true ;
  then false ;                         \ Nope, return false.

: quit ( -- )
  line-buffer adup a> 1 4+1 input-push \ Push pointer to word-buffer.
  0 c! adrop                           \ Store 0 there, so that we read a line.
  {
    get-word                           \ Read a word.
    word-buffer c@ adrop if            \ Not an empty word?
      word-buffer try-number if0       \ Try to compile as a number.
        word-buffer lookup             \ On error, try to interpret the word.
        try-word
      then
    then
  }

entry:
: main  init-pointers logo quit save-all exit-vm ;

[[ save flash.bin
bye

